@@ -1,7 +1,8 @@
 /*
     xattr - a python module for manipulating filesystem extended attributes
 
-    Copyright (C) 2002, 2003, 2006, 2008, 2012, 2013 Iustin Pop <iusty@k1024.org>
+    Copyright (C) 2002, 2003, 2006, 2008, 2012, 2013, 2015
+      Iustin Pop <iustin@k1024.org>
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -173,15 +174,17 @@ static int merge_ns(const char *ns, const char *name,
                     const char **result, char **buf) {
     if(ns != NULL && *ns != '\0') {
         int cnt;
+        /* The value of new_size is related to/must be kept in-sync
+           with the format string below */
         size_t new_size = strlen(ns) + 1 + strlen(name) + 1;
         if((*buf = PyMem_Malloc(new_size)) == NULL) {
             PyErr_NoMemory();
             return -1;
         }
         cnt = snprintf(*buf, new_size, "%s.%s", ns, name);
-        if(cnt > new_size || cnt < 0) {
+        if((size_t) cnt >= new_size || cnt < 0) {
             PyErr_SetString(PyExc_ValueError,
-                            "can't format the attribute name");
+                            "unexpected: can't format the attribute name");
             PyMem_Free(*buf);
             return -1;
         }
@@ -275,7 +278,8 @@ pygetxattr(PyObject *self, PyObject *args)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *res;
 
     /* Parse the arguments */
@@ -283,36 +287,38 @@ pygetxattr(PyObject *self, PyObject *args)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _get_obj(&tgt, attrname, NULL, 0)) == -1) {
+    if((nalloc_s = _get_obj(&tgt, attrname, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the attribute value */
     if((nret = _get_obj(&tgt, attrname, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Create the string which will hold the result */
     res = PyBytes_FromStringAndSize(buf, nret);
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -352,7 +358,8 @@ xattr_get(PyObject *self, PyObject *args, PyObject *keywds)
     const char *fullname;
     char *buf;
     const char *ns = NULL;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *res;
     static char *kwlist[] = {"item", "name", "nofollow", "namespace", NULL};
 
@@ -362,43 +369,45 @@ xattr_get(PyObject *self, PyObject *args, PyObject *keywds)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &fullname, &namebuf) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_tgt;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _get_obj(&tgt, fullname, NULL, 0)) == -1) {
+    if((nalloc_s = _get_obj(&tgt, fullname, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_name_buf;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freenamebuf;
+        goto free_name_buf;
     }
 
     /* Now retrieve the attribute value */
     if((nret = _get_obj(&tgt, fullname, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Create the string which will hold the result */
     res = PyBytes_FromStringAndSize(buf, nret);
 
     /* Free the buffers, they are no longer needed */
- freebuf:
+ free_buf:
     PyMem_Free(buf);
- freenamebuf:
+ free_name_buf:
     PyMem_Free(namebuf);
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -449,9 +458,10 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     PyObject *myarg, *res;
     int nofollow=0;
     const char *ns = NULL;
-    char *buf_list, *buf_val;
+    char *buf_list, *buf_val, *buf_val_tmp;
     const char *s;
-    ssize_t nalloc, nlist, nval;
+    ssize_t nalloc_s, nlist, nval_s;
+    size_t nalloc, nval;
     PyObject *mylist;
     target_t tgt;
     static char *kwlist[] = {"item", "nofollow", "namespace", NULL};
@@ -466,22 +476,24 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     /* Compute first the list of attributes */
 
     /* Find out the needed size of the buffer for the attribute list */
-    nalloc = _list_obj(&tgt, NULL, 0);
+    nalloc_s = _list_obj(&tgt, NULL, 0);
 
-    if(nalloc == -1) {
+    if(nalloc_s == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
-    if(nalloc == 0) {
+    if(nalloc_s == 0) {
         res = PyList_New(0);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf_list = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
@@ -517,19 +529,33 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
         /* Now retrieve the attribute value */
         missing = 0;
         while(1) {
-            nval = _get_obj(&tgt, s, buf_val, nalloc);
+            nval_s = _get_obj(&tgt, s, buf_val, nalloc);
 
-            if(nval == -1) {
+            if(nval_s == -1) {
                 if(errno == ERANGE) {
-                    nval = _get_obj(&tgt, s, NULL, 0);
-                    if((buf_val = PyMem_Realloc(buf_val, nval)) == NULL) {
+                    ssize_t realloc_size_s = _get_obj(&tgt, s, NULL, 0);
+                    /* ERANGE + proper size should not fail, but it
+                       still can, so let's check first */
+                    if(realloc_size_s == -1) {
+                      res = PyErr_SetFromErrno(PyExc_IOError);
+                      Py_DECREF(mylist);
+                      goto free_buf_val;
+                    }
+                    size_t realloc_size = (size_t) realloc_size_s;
+                    if((buf_val_tmp = PyMem_Realloc(buf_val, realloc_size))
+                       == NULL) {
                         res = PyErr_NoMemory();
                         Py_DECREF(mylist);
-                        goto free_buf_list;
+                        goto free_buf_val;
                     }
-                    nalloc = nval;
+                    buf_val = buf_val_tmp;
+                    nalloc = realloc_size;
                     continue;
-                } else if(errno == ENODATA || errno == ENOATTR) {
+                } else if(
+#ifdef ENODATA
+                          errno == ENODATA ||
+#endif
+                          errno == ENOATTR) {
                     /* this attribute has gone away since we queried
                        the attribute list */
                     missing = 1;
@@ -539,7 +565,9 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
                    don't know how to handle nicely, so we abort */
                 Py_DECREF(mylist);
                 res = PyErr_SetFromErrno(PyExc_IOError);
-                goto freebufval;
+                goto free_buf_val;
+            } else {
+              nval = (size_t) nval_s;
             }
             break;
         }
@@ -553,7 +581,7 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
         if (my_tuple == NULL) {
           Py_DECREF(mylist);
           res = NULL;
-          goto freebufval;
+          goto free_buf_val;
         }
         PyList_Append(mylist, my_tuple);
         Py_DECREF(my_tuple);
@@ -562,13 +590,13 @@ get_all(PyObject *self, PyObject *args, PyObject *keywds)
     /* Successful exit */
     res = mylist;
 
- freebufval:
+ free_buf_val:
     PyMem_Free(buf_val);
 
  free_buf_list:
     PyMem_Free(buf_list);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
 
     /* Return the result */
@@ -604,18 +632,28 @@ pysetxattr(PyObject *self, PyObject *args)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf = NULL;
-    int bufsize;
+    Py_ssize_t bufsize_s;
+    size_t bufsize;
     int nret;
     int flags = 0;
     target_t tgt;
 
     /* Parse the arguments */
     if (!PyArg_ParseTuple(args, "Oetet#|ii", &myarg, NULL, &attrname,
-                          NULL, &buf, &bufsize, &flags, &nofollow))
+                          NULL, &buf, &bufsize_s, &flags, &nofollow))
         return NULL;
+
+    if (bufsize_s < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                        "negative value size?!");
+        res = NULL;
+        goto free_arg;
+    }
+    bufsize = (size_t) bufsize_s;
+
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Set the attribute's value */
@@ -625,13 +663,13 @@ pysetxattr(PyObject *self, PyObject *args)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
     PyMem_Free(buf);
 
@@ -670,7 +708,8 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
     int nofollow = 0;
     char *attrname = NULL;
     char *buf = NULL;
-    int bufsize;
+    Py_ssize_t bufsize_s;
+    size_t bufsize;
     int nret;
     int flags = 0;
     target_t tgt;
@@ -683,16 +722,25 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
     /* Parse the arguments */
     if (!PyArg_ParseTupleAndKeywords(args, keywds, "Oetet#|ii" BYTES_CHAR,
                                      kwlist, &myarg, NULL, &attrname, NULL,
-                                     &buf, &bufsize, &flags, &nofollow, &ns))
+                                     &buf, &bufsize_s, &flags, &nofollow, &ns))
         return NULL;
+
+    if (bufsize_s < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                        "negative value size?!");
+        res = NULL;
+        goto free_arg;
+    }
+    bufsize = (size_t) bufsize_s;
+
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &full_name, &newname) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Set the attribute's value */
@@ -704,13 +752,13 @@ xattr_set(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
     PyMem_Free(buf);
 
@@ -747,7 +795,7 @@ pyremovexattr(PyObject *self, PyObject *args)
 
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Remove the attribute */
@@ -757,13 +805,13 @@ pyremovexattr(PyObject *self, PyObject *args)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -809,12 +857,12 @@ xattr_remove(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     if(merge_ns(ns, attrname, &full_name, &name_buf) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Remove the attribute */
@@ -826,13 +874,13 @@ xattr_remove(PyObject *self, PyObject *args, PyObject *keywds)
 
     if(nret == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freearg;
+        goto free_arg;
     }
 
     Py_INCREF(Py_None);
     res = Py_None;
 
- freearg:
+ free_arg:
     PyMem_Free(attrname);
 
     /* Return the result */
@@ -856,7 +904,8 @@ pylistxattr(PyObject *self, PyObject *args)
 {
     char *buf;
     int nofollow=0;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *myarg;
     PyObject *mylist;
     Py_ssize_t nattrs;
@@ -870,26 +919,28 @@ pylistxattr(PyObject *self, PyObject *args)
         return NULL;
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _list_obj(&tgt, NULL, 0)) == -1) {
+    if((nalloc_s = _list_obj(&tgt, NULL, 0)) == -1) {
         mylist = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
-    if(nalloc == 0) {
+    if(nalloc_s == 0) {
         mylist = PyList_New(0);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         mylist = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
     if((nret = _list_obj(&tgt, buf, nalloc)) == -1) {
         mylist = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Compute the number of attributes in the list */
@@ -900,7 +951,7 @@ pylistxattr(PyObject *self, PyObject *args)
     /* Create the list which will hold the result */
     mylist = PyList_New(nattrs);
     if(mylist == NULL)
-      goto freebuf;
+      goto free_buf;
 
     /* Create and insert the attributes as strings in the list */
     for(s = buf, nattrs = 0; s - buf < nret; s += strlen(s) + 1) {
@@ -908,17 +959,17 @@ pylistxattr(PyObject *self, PyObject *args)
         if(item == NULL) {
             Py_DECREF(mylist);
             mylist = NULL;
-            goto freebuf;
+            goto free_buf;
         }
         PyList_SET_ITEM(mylist, nattrs, item);
         nattrs++;
     }
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
 
     /* Return the result */
@@ -956,7 +1007,8 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
 {
     char *buf;
     int nofollow = 0;
-    ssize_t nalloc, nret;
+    ssize_t nalloc_s, nret;
+    size_t nalloc;
     PyObject *myarg;
     PyObject *res;
     const char *ns = NULL;
@@ -971,30 +1023,32 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
         return NULL;
     if(convert_obj(myarg, &tgt, nofollow) < 0) {
         res = NULL;
-        goto freearg;
+        goto free_arg;
     }
 
     /* Find out the needed size of the buffer */
-    if((nalloc = _list_obj(&tgt, NULL, 0)) == -1) {
+    if((nalloc_s = _list_obj(&tgt, NULL, 0)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freetgt;
+        goto free_tgt;
     }
 
-    if(nalloc == 0) {
+    if(nalloc_s == 0) {
         res = PyList_New(0);
-        goto freetgt;
+        goto free_tgt;
     }
 
+    nalloc = (size_t) nalloc_s;
+
     /* Try to allocate the memory, using Python's allocator */
     if((buf = PyMem_Malloc(nalloc)) == NULL) {
         res = PyErr_NoMemory();
-        goto freetgt;
+        goto free_tgt;
     }
 
     /* Now retrieve the list of attributes */
     if((nret = _list_obj(&tgt, buf, nalloc)) == -1) {
         res = PyErr_SetFromErrno(PyExc_IOError);
-        goto freebuf;
+        goto free_buf;
     }
 
     /* Compute the number of attributes in the list */
@@ -1005,7 +1059,7 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
     /* Create the list which will hold the result */
     res = PyList_New(nattrs);
     if(res == NULL)
-        goto freebuf;
+        goto free_buf;
 
     /* Create and insert the attributes as strings in the list */
     for(s = buf, nattrs = 0; s - buf < nret; s += strlen(s) + 1) {
@@ -1015,20 +1069,20 @@ xattr_list(PyObject *self, PyObject *args, PyObject *keywds)
             if(item == NULL) {
                 Py_DECREF(res);
                 res = NULL;
-                goto freebuf;
+                goto free_buf;
             }
             PyList_SET_ITEM(res, nattrs, item);
             nattrs++;
         }
     }
 
- freebuf:
+ free_buf:
     /* Free the buffer, now it is no longer needed */
     PyMem_Free(buf);
 
- freetgt:
+ free_tgt:
     free_tgt(&tgt);
- freearg:
+ free_arg:
 
     /* Return the result */
     return res;
